@@ -36,6 +36,7 @@ import (
 	"time"
 
 	apko_types "chainguard.dev/apko/pkg/build/types"
+	"github.com/github/go-spdx/v2/spdxexp"
 	purl "github.com/package-url/packageurl-go"
 
 	"chainguard.dev/melange/pkg/sbom"
@@ -441,7 +442,9 @@ type Copyright struct {
 }
 
 // LicenseExpression returns an SPDX license expression formed from the data in
-// the copyright structs found in the conf. It's a simple OR for now.
+// the copyright structs found in the conf. It's a simple AND for now.
+// Invalid SPDX license identifiers are converted to LicenseRef-<sanitized-name>
+// format for SPDX compliance.
 func (p Package) LicenseExpression() string {
 	licenseExpression := ""
 	if p.Copyright == nil {
@@ -451,26 +454,99 @@ func (p Package) LicenseExpression() string {
 		if licenseExpression != "" {
 			licenseExpression += " AND "
 		}
-		licenseExpression += cp.License
+		license := normalizeLicenseID(cp.License)
+		licenseExpression += license
 	}
 	return licenseExpression
 }
 
+// normalizeLicenseID checks if a license identifier is a valid SPDX license.
+// If valid, it returns the license as-is. If invalid, it returns a LicenseRef-
+// formatted identifier that is SPDX compliant.
+func normalizeLicenseID(license string) string {
+	if license == "" {
+		return ""
+	}
+
+	if isValidSPDXLicense(license) {
+		return license
+	}
+
+	// Convert invalid license to LicenseRef format
+	return toLicenseRef(license)
+}
+
+// toLicenseRef converts a license name to a valid SPDX LicenseRef identifier.
+// SPDX LicenseRef identifiers must match the pattern: LicenseRef-[a-zA-Z0-9.-]+
+func toLicenseRef(license string) string {
+	// Sanitize the license name to only contain valid characters
+	var sb strings.Builder
+	for _, r := range license {
+		switch {
+		case (r >= 'a' && r <= 'z') || (r >= 'A' && r <= 'Z') || (r >= '0' && r <= '9') || r == '.' || r == '-':
+			sb.WriteRune(r)
+		case r == ' ' || r == '_':
+			sb.WriteRune('-')
+		default:
+			// Skip invalid characters
+		}
+	}
+
+	sanitized := sb.String()
+	if sanitized == "" {
+		// If nothing remains after sanitization, use a hash
+		h := sha256.Sum256([]byte(license))
+		sanitized = hex.EncodeToString(h[:8])
+	}
+
+	return "LicenseRef-" + sanitized
+}
+
+// isValidSPDXLicense checks if a license identifier is a valid SPDX license.
+func isValidSPDXLicense(license string) bool {
+	if license == "" {
+		return false
+	}
+	valid, _ := spdxexp.ValidateLicenses([]string{license})
+	return valid
+}
+
 // LicensingInfos looks at the `Package.Copyright[].LicensePath` fields of the
 // parsed build configuration for the package. If this value has been set,
 // LicensingInfos opens the file at this path from the build's workspace
 // directory, and reads in the license content. LicensingInfos then returns a
-// map of the `Copyright.License` field to the string content of the file from
-// `.LicensePath`.
-func (p Package) LicensingInfos(workspaceDir string) (map[string]string, error) {
+// map of the license identifier (using LicenseRef format for non-SPDX licenses)
+// to the string content of the file from `.LicensePath`.
+//
+// For licenses that are not valid SPDX identifiers, the extracted text will
+// include the original license name if no LicensePath is specified. A warning
+// is logged in this case to encourage providing the full license text.
+func (p Package) LicensingInfos(ctx context.Context, workspaceDir string) (map[string]string, error) {
+	log := clog.FromContext(ctx)
 	licenseInfos := make(map[string]string)
 	for _, cp := range p.Copyright {
+		license := cp.License
+		id := normalizeLicenseID(license)
+
 		if cp.LicensePath != "" {
-			content, err := os.ReadFile(filepath.Join(workspaceDir, cp.LicensePath)) // #nosec G304 - Reading license file from build workspace
+			// Clean and localize the path
+			cleanPath := filepath.Clean(cp.LicensePath)
+			localPath, err := filepath.Localize(cleanPath)
+			if err != nil {
+				return nil, fmt.Errorf("invalid license-path %q: %w", cp.LicensePath, err)
+			}
+			fullPath := filepath.Join(workspaceDir, localPath)
+
+			content, err := os.ReadFile(fullPath) // #nosec G304 - Reading license file from build workspace
 			if err != nil {
 				return nil, fmt.Errorf("failed to read licensepath %q: %w", cp.LicensePath, err)
 			}
-			licenseInfos[cp.License] = string(content)
+			licenseInfos[id] = string(content)
+		} else if strings.HasPrefix(id, "LicenseRef-") {
+			// For non-SPDX licenses without a license path, include the original
+			// license name as the extracted text so consumers know what it refers to
+			log.Warnf("non-SPDX license %q used without license-path; consider adding license-path with full license text for SBOM compliance", license)
+			licenseInfos[id] = fmt.Sprintf("Non-SPDX License: %s", license)
 		}
 	}
 	return licenseInfos, nil
@@ -653,6 +729,7 @@ func getGitSBOMPackage(repo, tag, expectedCommit string, idComponents []string,
 			Namespace:        namespace,
 			PURL:             pu,
 			DownloadLocation: downloadLocation,
+			PrimaryPurpose:   "SOURCE",
 		}, nil
 	}
 
@@ -705,7 +782,7 @@ func (p Pipeline) SBOMPackageForUpstreamSource(licenseDeclared, supplier string,
 			return nil, err
 		}
 
-		idComponents := []string{pkgName, pkgVersion}
+		idComponents := []string{"Source", pkgName, pkgVersion}
 		if uniqueID != "" {
 			idComponents = append(idComponents, uniqueID)
 		}
@@ -718,6 +795,7 @@ func (p Pipeline) SBOMPackageForUpstreamSource(licenseDeclared, supplier string,
 			Checksums:        checksums,
 			PURL:             pu,
 			DownloadLocation: args["download_url"],
+			PrimaryPurpose:   "SOURCE",
 		}, nil
 
 	case "git-checkout":
@@ -735,7 +813,7 @@ func (p Pipeline) SBOMPackageForUpstreamSource(licenseDeclared, supplier string,
 			s = strings.TrimPrefix(s, "http://")
 			return s
 		}()
-		for _, component := range []string{repoCleaned, branch, tag, expectedCommit} {
+		for _, component := range []string{"Source", repoCleaned, branch, tag, expectedCommit} {
 			if component != "" {
 				idComponents = append(idComponents, component)
 			}
@@ -1102,7 +1180,7 @@ type ConfigurationParsingOption func(*configOptions)
 
 type configOptions struct {
 	filesystem                  fs.FS
-	envFilePath                 string
+	envFilePaths                []string
 	cpu, cpumodel, memory, disk string
 	timeout                     time.Duration
 	commit                      string
@@ -1163,10 +1241,14 @@ func WithCommit(hash string) ConfigurationParsingOption {
 	}
 }
 
-// WithEnvFileForParsing set the paths from which to read an environment file.
-func WithEnvFileForParsing(path string) ConfigurationParsingOption {
+// WithEnvFilesForParsing set the paths from which to read environment files.
+func WithEnvFilesForParsing(paths []string) ConfigurationParsingOption {
 	return func(options *configOptions) {
-		options.envFilePath = path
+		for _, path := range paths {
+			if path != "" {
+				options.envFilePaths = append(options.envFilePaths, path)
+			}
+		}
 	}
 }
 
@@ -1197,7 +1279,7 @@ func buildConfigMap(cfg *Configuration) map[string]string {
 }
 
 func replacerFromMap(with map[string]string) *strings.Replacer {
-	replacements := []string{}
+	replacements := make([]string, 0, len(with)*2)
 	for k, v := range with {
 		replacements = append(replacements, k, v)
 	}
@@ -1603,7 +1685,7 @@ func ParseConfiguration(ctx context.Context, configurationFilePath string, opts
 	}
 
 	// Merge environment file if needed.
-	if envFile := options.envFilePath; envFile != "" {
+	for _, envFile := range options.envFilePaths {
 		envMap, err := godotenv.Read(envFile)
 		if err != nil {
 			return nil, fmt.Errorf("loading environment file: %w", err)
@@ -1622,9 +1704,14 @@ func ParseConfiguration(ctx context.Context, configurationFilePath string, opts
 	}
 
 	const (
-		defaultEnvVarHOME       = "/home/build"
-		defaultEnvVarGOPATH     = "/home/build/.cache/go"
-		defaultEnvVarGOMODCACHE = "/var/cache/melange/gomodcache"
+		defaultEnvVarHOME             = "/home/build"
+		defaultEnvVarGOPATH           = "/home/build/.cache/go"
+		defaultEnvVarGOMODCACHE       = "/var/cache/melange/gomodcache"
+		defaultEnvVarUVCACHEDIR       = "/var/cache/melange/uv"
+		defaultEnvVarPIPCACHEDIR      = "/var/cache/melange/pip"
+		defaultEnvVarCOMPOSERCACHEDIR = "/var/cache/melange/composer"
+		defaultEnvVarNPMCACHE         = "/var/cache/melange/npm"
+		defaultEnvVarCARGOHOME        = "/var/cache/melange/cargo"
 	)
 
 	setIfEmpty := func(key, value string) {
@@ -1636,6 +1723,11 @@ func ParseConfiguration(ctx context.Context, configurationFilePath string, opts
 	setIfEmpty("HOME", defaultEnvVarHOME)
 	setIfEmpty("GOPATH", defaultEnvVarGOPATH)
 	setIfEmpty("GOMODCACHE", defaultEnvVarGOMODCACHE)
+	setIfEmpty("UV_CACHE_DIR", defaultEnvVarUVCACHEDIR)
+	setIfEmpty("PIP_CACHE_DIR", defaultEnvVarPIPCACHEDIR)
+	setIfEmpty("COMPOSER_CACHE_DIR", defaultEnvVarCOMPOSERCACHEDIR)
+	setIfEmpty("npm_config_cache", defaultEnvVarNPMCACHE)
+	setIfEmpty("CARGO_HOME", defaultEnvVarCARGOHOME)
 
 	if err := cfg.applySubstitutionsForProvides(); err != nil {
 		return nil, err
