@@ -23,6 +23,7 @@ import (
 	"crypto/ed25519"
 	"crypto/elliptic"
 	"crypto/rand"
+	"crypto/sha256"
 	_ "embed"
 	"encoding/base64"
 	"encoding/pem"
@@ -35,6 +36,7 @@ import (
 	"os/signal"
 	"os/user"
 	"path/filepath"
+	"regexp"
 	"runtime"
 	"strconv"
 	"strings"
@@ -388,6 +390,7 @@ func (bw *qemu) TerminatePod(ctx context.Context, cfg *Config) error {
 	defer os.Remove(cfg.Disk)
 	defer os.Remove(cfg.SSHHostKey)
 	defer secureDelete(ctx, cfg.InitramfsPath)
+	defer stopVirtiofsd(ctx, cfg)
 
 	clog.FromContext(ctx).Info("qemu: sending shutdown signal")
 	err := sendSSHCommand(ctx,
@@ -444,6 +447,13 @@ func (bw *qemu) WorkspaceTar(ctx context.Context, cfg *Config, extraFiles []stri
 	}
 	defer outFile.Close()
 
+	errFile_path := filepath.Join(cfg.WorkspaceDir, ".error-file")
+	errFile, err := os.Create(errFile_path)
+	if err != nil {
+		return nil, err
+	}
+	defer errFile.Close()
+
 	clog.FromContext(ctx).Infof("fetching remote workspace")
 	// work around missing scp (needs openssh-sftp package), we just tar the file
 	// and pipe the output to our local file. It is potentially slower, but being
@@ -459,26 +469,25 @@ func (bw *qemu) WorkspaceTar(ctx context.Context, cfg *Config, extraFiles []stri
 		retrieveCommand += " -T extrafiles.txt"
 	}
 
-	log := clog.FromContext(ctx)
-	stderr := logwriter.New(log.Debug)
 	err = sendSSHCommand(ctx,
 		cfg.SSHControlClient,
 		cfg,
 		nil,
-		stderr,
+		errFile,
 		outFile,
 		false,
 		[]string{"sh", "-c", retrieveCommand},
 	)
 	if err != nil {
-		var buf bytes.Buffer
-		_, cerr := io.Copy(&buf, outFile)
-		if cerr != nil {
-			clog.FromContext(ctx).Errorf("failed to tar workspace: %v", cerr)
-			return nil, cerr
+		savederr := err
+		outErr, err := os.ReadFile(errFile_path)
+		if err == nil {
+			clog.FromContext(ctx).Errorf("failed to tar workspace:\n%s", string(outErr))
+		} else {
+			clog.FromContext(ctx).Errorf("failed to tar workspace; failed to obtain error output")
 		}
-		clog.FromContext(ctx).Errorf("failed to tar workspace: %v", buf.String())
-		return nil, err
+
+		return nil, savederr
 	}
 
 	return os.Open(outFile.Name())
@@ -588,17 +597,38 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		return err
 	}
 
-	baseargs := []string{}
-	bios := false
-	useVM := false
-	if qemuVM, ok := os.LookupEnv("QEMU_USE_MICROVM"); ok {
-		if val, err := strconv.ParseBool(qemuVM); err == nil {
-			useVM = val
+	// Set up virtiofs for cache directory if enabled and available
+	if cfg.CacheDir != "" {
+		// Ensure the cachedir exists
+		if err := os.MkdirAll(cfg.CacheDir, 0o755); err != nil {
+			return fmt.Errorf("failed to create shared cachedir: %w", err)
+		}
+
+		var err error
+		cfg.VirtiofsEnabled, err = useVirtiofs()
+		if err != nil {
+			return err
+		}
+
+		if cfg.VirtiofsEnabled {
+			// Generate a random socket path
+			var randBytes [8]byte
+			if _, err := rand.Read(randBytes[:]); err != nil {
+				return fmt.Errorf("failed to generate random bytes for socket path: %w", err)
+			}
+			cfg.VirtiofsdSocketPath = filepath.Join(os.TempDir(), fmt.Sprintf("melange-virtiofsd-%x.sock", randBytes))
+
+			virtiofsdCmd, err := startVirtiofsd(ctx, cfg)
+			if err != nil {
+				return fmt.Errorf("failed to start virtiofsd: %w", err)
+			}
+			cfg.VirtiofsdPID = virtiofsdCmd.Process.Pid
 		}
 	}
 
-	kernelConsole := "console=hvc0"
+	baseargs := []string{}
 
+	kernelConsole := ""
 	// If the log level is debug, then crank up the logging.
 	// Otherwise, use quiet mode.
 	if log.Enabled(ctx, slog.LevelDebug) {
@@ -607,44 +637,33 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		kernelConsole += " quiet"
 	}
 
-	serialArgs := []string{
-		"-device", "virtio-serial-pci,id=virtio-serial0",
-		"-chardev", "stdio,id=charconsole0",
-		"-device", "virtconsole,chardev=charconsole0,id=console0",
-	}
-	if useVM {
-		// load microvm profile and bios, shave some milliseconds from boot
-		// using this will make a complete boot->initrd (with working network) In ~700ms
-		// instead of ~900ms.
-		for _, p := range []string{
-			"/usr/share/qemu/bios-microvm.bin",
-			"/usr/share/seabios/bios-microvm.bin",
-		} {
-			if _, err := os.Stat(p); err == nil && cfg.Arch.ToAPK() != "aarch64" {
-				// only enable pcie for network, enable RTC for kernel, disable i8254PIT, i8259PIC and serial port
-				baseargs = append(baseargs, "-machine", "microvm,rtc=on,pcie=on,pit=off,pic=off,isa-serial=on")
-				baseargs = append(baseargs, "-bios", p)
-				// microvm in qemu any version tested will not send hvc0/virtconsole to stdout
-				kernelConsole = "console=ttyS0"
-				serialArgs = []string{"-serial", "stdio"}
-				bios = true
-				break
-			}
+	// Only enable console on debug runs.
+	// Spare some boot time and memory
+	serialArgs := []string{}
+	if log.Enabled(ctx, slog.LevelDebug) {
+		kernelConsole = "console=hvc0"
+		serialArgs = []string{
+			"-device", "virtio-serial-pci,id=virtio-serial0,max_ports=2",
+			"-chardev", "stdio,id=charconsole0",
+			"-device", "virtconsole,chardev=charconsole0,id=console0",
 		}
 	}
 
-	// we need to fall back to -machine virt if no microVM BIOS was found (or QEMU_USE_MICROVM is false)
-	if !bios {
-		// aarch64 supports virt machine type, let's use that if we're on it, else
-		// if we're on x86 arch, but without microvm machine type, let's go to q35
-		switch cfg.Arch.ToAPK() {
-		case "aarch64":
-			baseargs = append(baseargs, "-machine", "virt")
-		case "x86_64":
-			baseargs = append(baseargs, "-machine", "q35")
-		default:
-			return fmt.Errorf("unknown architecture: %s", cfg.Arch.ToAPK())
-		}
+	// Helper to add memory-backend suffix for virtiofs shared memory
+	machineMemorySuffix := ""
+	if cfg.VirtiofsEnabled {
+		machineMemorySuffix = ",memory-backend=mem"
+	}
+
+	// aarch64 supports virt machine type, let's use that if we're on it, else
+	// if we're on x86 arch, but without microvm machine type, let's go to q35
+	switch cfg.Arch.ToAPK() {
+	case "aarch64":
+		baseargs = append(baseargs, "-machine", "virt"+machineMemorySuffix)
+	case "x86_64":
+		baseargs = append(baseargs, "-machine", "q35,hpet=off,smm=off,pit=off,i8042=off,sata=off,smbus=off,usb=off,vmport=off,graphics=off"+machineMemorySuffix)
+	default:
+		return fmt.Errorf("unknown architecture: %s", cfg.Arch.ToAPK())
 	}
 
 	// default to use 85% of available memory, if a mem limit is set, respect it.
@@ -659,7 +678,15 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 			mem = memKb
 		}
 	}
-	baseargs = append(baseargs, "-m", fmt.Sprintf("%dk", mem))
+	// Memory configuration - virtiofs requires shared memory backend
+	if cfg.VirtiofsEnabled {
+		// Round up memory to MiB boundary for memory-backend-memfd alignment
+		memMiB := (mem + 1023) / 1024 * 1024
+		baseargs = append(baseargs, "-m", fmt.Sprintf("%dk", memMiB))
+		baseargs = append(baseargs, "-object", fmt.Sprintf("memory-backend-memfd,id=mem,size=%dM,share=on", memMiB/1024))
+	} else {
+		baseargs = append(baseargs, "-m", fmt.Sprintf("%dk", mem))
+	}
 
 	// default to use all CPUs, if a cpu limit is set, respect it.
 	nproc := runtime.NumCPU()
@@ -699,19 +726,23 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		baseargs = append(baseargs, "-cpu", "host")
 	}
 
+	cmdlineVar, ok := os.LookupEnv("QEMU_KERNEL_CMDLINE")
+	if ok {
+		clog.FromContext(ctx).Infof("qemu: QEMU_KERNEL_CMDLINE set to: %s", cmdlineVar)
+	} else {
+		cmdlineVar = ""
+	}
+
 	// ensure we disable unneeded devices, this is less needed if we use microvm machines
 	// but still useful otherwise
 	baseargs = append(baseargs, "-display", "none")
 	baseargs = append(baseargs, "-no-reboot")
 	baseargs = append(baseargs, "-no-user-config")
-	baseargs = append(baseargs, "-nographic")
 	baseargs = append(baseargs, "-nodefaults")
-	baseargs = append(baseargs, "-parallel", "none")
 	baseargs = append(baseargs, serialArgs...)
-	baseargs = append(baseargs, "-vga", "none")
 	// use -netdev + -device instead of -nic, as this is better supported by microvm machine type
 	baseargs = append(baseargs, "-netdev", "user,id=id1,hostfwd=tcp:"+cfg.SSHAddress+"-:22,hostfwd=tcp:"+cfg.SSHControlAddress+"-:2223")
-	baseargs = append(baseargs, "-device", "virtio-net-pci,netdev=id1")
+	baseargs = append(baseargs, "-device", "virtio-net-pci,netdev=id1,romfile=")
 	// add random generator via pci, improve ssh startup time
 	baseargs = append(baseargs, "-device", "virtio-rng-pci,rng=rng0", "-object", "rng-random,filename=/dev/urandom,id=rng0")
 	// panic=-1 ensures that if the init fails, we immediately exit the machine
@@ -719,7 +750,14 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	sshkey := base64.StdEncoding.EncodeToString(pubKey)
 
 	// Build kernel command line arguments
-	kernelArgs := kernelConsole + " nomodeset random.trust_cpu=on panic=-1 sshkey=" + sshkey + " melange_qemu_runner=1"
+	kernelArgs := kernelConsole +
+		" nomodeset random.trust_cpu=on" +
+		" noapic nomodules 8250.nr_uarts=0 tsc=reliable" +
+		" no_timer_check cryptomgr.notests" +
+		" rcupdate.rcu_expedited=1 panic=-1" +
+		" pci=lastbus=0 ftrace=off swiotlb=noforce edd=off " +
+		cmdlineVar + " sshkey=" + sshkey +
+		" melange_qemu_runner=1"
 
 	// Check for TESTING environment variable and pass it to microvm-init
 	// TESTING must be a number (0 for disabled, non-zero for enabled)
@@ -742,12 +780,19 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	baseargs = append(baseargs, "-device", "virtio-9p-pci,id=fs100,fsdev=fsdev100,mount_tag=defaultshare")
 
 	if cfg.CacheDir != "" {
-		baseargs = append(baseargs, "-fsdev", "local,security_model=mapped,id=fsdev101,readonly=on,path="+cfg.CacheDir)
-		baseargs = append(baseargs, "-device", "virtio-9p-pci,id=fs101,fsdev=fsdev101,mount_tag=melange_cache")
-
-		// ensure the cachedir exists
-		if err := os.MkdirAll(cfg.CacheDir, 0o755); err != nil {
-			return fmt.Errorf("failed to create shared cachedir: %w", err)
+		if cfg.VirtiofsEnabled {
+			log.Info("qemu: using virtiofs for cache directory (read-write)")
+			// Chardev for socket communication
+			baseargs = append(baseargs,
+				"-chardev",
+				fmt.Sprintf("socket,id=char_cache,path=%s", cfg.VirtiofsdSocketPath))
+			// vhost-user-fs-pci device
+			baseargs = append(baseargs, "-device",
+				"vhost-user-fs-pci,queue-size=1024,chardev=char_cache,tag=melange_cache")
+		} else {
+			log.Info("qemu: using 9p for cache directory (read-only with overlay)")
+			baseargs = append(baseargs, "-fsdev", "local,security_model=mapped,id=fsdev101,readonly=on,path="+cfg.CacheDir)
+			baseargs = append(baseargs, "-device", "virtio-9p-pci,id=fs101,fsdev=fsdev101,mount_tag=melange_cache")
 		}
 	}
 
@@ -793,17 +838,12 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 
 	// append raw disk, init will take care of formatting it if present.
 	baseargs = append(baseargs, "-object", "iothread,id=io1")
-	baseargs = append(baseargs, "-device", "virtio-blk-pci,drive=disk0,iothread=io1,packed=on,num-queues="+fmt.Sprintf("%d", nproc/2))
-	if runtime.GOOS == "linux" {
-		baseargs = append(baseargs, "-drive", "if=none,id=disk0,cache=unsafe,cache.direct=on,format=raw,aio=native,file="+diskFile)
-	}
-	if runtime.GOOS == "darwin" {
-		baseargs = append(baseargs, "-drive", "if=none,id=disk0,cache=unsafe,format=raw,aio=threads,file="+diskFile)
-	}
+	baseargs = append(baseargs, "-device", "virtio-blk-pci,drive=disk0,iothread=io1,packed=on,num-queues="+fmt.Sprintf("%d", max(1, nproc/2)))
+	baseargs = append(baseargs, "-drive", "if=none,id=disk0,cache=unsafe,format=raw,aio=threads,file="+diskFile)
 
 	// append the rootfs tar.gz, init will take care of populating the disk with it
 	baseargs = append(baseargs, "-object", "iothread,id=io2")
-	baseargs = append(baseargs, "-device", "virtio-blk-pci,drive=image.tar,iothread=io2,packed=on,num-queues="+fmt.Sprintf("%d", nproc/2))
+	baseargs = append(baseargs, "-device", "virtio-blk-pci,drive=image.tar,iothread=io2,packed=on,num-queues="+fmt.Sprintf("%d", max(1, nproc/2)))
 	baseargs = append(baseargs, "-blockdev", "driver=raw,node-name=image.tar,file.driver=file,file.filename="+cfg.ImgRef)
 
 	// qemu-system-x86_64 or qemu-system-aarch64...
@@ -821,11 +861,12 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	if err := qemuCmd.Start(); err != nil {
 		defer os.Remove(cfg.ImgRef)
 		defer os.Remove(cfg.Disk)
+		defer stopVirtiofsd(ctx, cfg)
 		return fmt.Errorf("qemu: failed to start qemu command: %w", err)
 	}
 
-	logCtx, cancel := context.WithCancel(ctx)
-	defer cancel()
+	logCtx, cancel := context.WithCancelCause(ctx)
+	defer cancel(fmt.Errorf("normal cleanup"))
 
 	go func() {
 		defer outRead.Close()
@@ -863,7 +904,7 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 
 	clog.FromContext(ctx).Info("qemu: waiting for SSH")
 	go func() {
-		// one-min timeout with a 500ms sleep
+		// 100ms timeout with a 100ms sleep
 		retries := 60
 		try := 0
 		for try < retries {
@@ -872,7 +913,7 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 			}
 
 			try++
-			time.Sleep(time.Millisecond * 500)
+			time.Sleep(time.Millisecond * 100)
 
 			log.Debugf("qemu: waiting for ssh to come up, try %d of %d", try, retries)
 			err = checkSSHServer(cfg.SSHAddress)
@@ -883,6 +924,8 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 				log.Info(err.Error())
 			}
 		}
+
+		cancel(fmt.Errorf("ssh server never came up"))
 	}()
 
 	select {
@@ -893,11 +936,13 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	case err := <-qemuExit:
 		defer os.Remove(cfg.ImgRef)
 		defer os.Remove(cfg.Disk)
+		defer stopVirtiofsd(ctx, cfg)
 		return fmt.Errorf("qemu: VM exited unexpectedly: %w", err)
 	case <-ctx.Done():
 		defer os.Remove(cfg.ImgRef)
 		defer os.Remove(cfg.Disk)
-		return fmt.Errorf("qemu: context canceled while waiting for VM to start")
+		defer stopVirtiofsd(ctx, cfg)
+		return fmt.Errorf("qemu: context canceled while waiting for VM to start: %w", context.Cause(ctx))
 	}
 
 	err = getHostKey(ctx, cfg)
@@ -925,33 +970,52 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	clog.FromContext(ctx).Infof("qemu: running kernel version: %s", kv)
 
 	if cfg.CacheDir != "" {
-		clog.FromContext(ctx).Infof("qemu: setting up melange cachedir: %s", cfg.CacheDir)
-		setupMountCommand := fmt.Sprintf(
-			"mkdir -p %s %s /mount/upper /mount/work && "+
-				"chmod 1777 /mount/upper && "+
-				"mount -t 9p -o ro melange_cache %s && "+
-				"mount -t overlay overlay -o lowerdir=%s,upperdir=/mount/upper,workdir=/mount/work %s",
-			DefaultCacheDir,
-			filepath.Join("/mount", DefaultCacheDir),
-			DefaultCacheDir,
-			DefaultCacheDir,
-			filepath.Join("/mount", DefaultCacheDir),
-		)
-		if setupMountCommand != ": " {
-			err = sendSSHCommand(ctx,
-				cfg.SSHControlClient,
-				cfg,
-				nil,
-				stderr,
-				stdout,
-				false,
-				[]string{"sh", "-c", setupMountCommand},
+		var setupMountCommand string
+
+		if cfg.VirtiofsEnabled {
+			// Virtiofs: read-write mount at the chroot path
+			// The build runs chrooted at /mount/, so we mount at /mount/var/cache/melange
+			// which appears as /var/cache/melange inside the chroot
+			chrootCacheDir := filepath.Join("/mount", DefaultCacheDir)
+			clog.FromContext(ctx).Infof("qemu: setting up virtiofs cache mount (read-write): %s -> %s", cfg.CacheDir, chrootCacheDir)
+			setupMountCommand = fmt.Sprintf(
+				"mkdir -p %s && mount -t virtiofs melange_cache %s",
+				chrootCacheDir,
+				chrootCacheDir,
 			)
+		} else {
+			// 9p: readonly with overlay for write support
+			clog.FromContext(ctx).Infof("qemu: setting up 9p cache mount (read-only with overlay): %s", cfg.CacheDir)
+			setupMountCommand = fmt.Sprintf(
+				"mkdir -p %s %s /mount/upper /mount/work && "+
+					"chmod 1777 /mount/upper && "+
+					"mount -t 9p -o ro melange_cache %s && "+
+					"mount -t overlay overlay -o lowerdir=%s,upperdir=/mount/upper,workdir=/mount/work %s",
+				DefaultCacheDir,
+				filepath.Join("/mount", DefaultCacheDir),
+				DefaultCacheDir,
+				DefaultCacheDir,
+				filepath.Join("/mount", DefaultCacheDir),
+			)
+		}
+
+		err = sendSSHCommand(ctx,
+			cfg.SSHControlClient,
+			cfg,
+			nil,
+			stderr,
+			stdout,
+			false,
+			[]string{"sh", "-c", setupMountCommand},
+		)
+		if err != nil {
+			// Clean up virtiofsd on failure
+			if cfg.VirtiofsEnabled {
+				stopVirtiofsd(ctx, cfg)
+			}
+			err = qemuCmd.Process.Kill()
 			if err != nil {
-				err = qemuCmd.Process.Kill()
-				if err != nil {
-					return err
-				}
+				return err
 			}
 		}
 	}
@@ -1660,6 +1724,133 @@ func randomPortN() (int, error) {
 	return l.Addr().(*net.TCPAddr).Port, nil
 }
 
+// virtiofsdSearchPaths defines where to look for the virtiofsd binary.
+var virtiofsdSearchPaths = []string{
+	"/usr/libexec/virtiofsd",  // Fedora, RHEL
+	"/usr/lib/qemu/virtiofsd", // Debian, Ubuntu
+	"virtiofsd",               // In PATH
+}
+
+// isVirtiofsdAvailable checks if virtiofsd is installed and returns its path.
+// If QEMU_VIRTIOFS_PATH is set, it will look for virtiofsd in that directory.
+func isVirtiofsdAvailable() (string, bool) {
+	// Check for user-specified directory first (useful for macOS/brew or custom installs)
+	if customDir, ok := os.LookupEnv("QEMU_VIRTIOFS_PATH"); ok && customDir != "" {
+		customPath := filepath.Join(customDir, "virtiofsd")
+		if _, err := os.Stat(customPath); err == nil {
+			return customPath, true
+		}
+	}
+	for _, path := range virtiofsdSearchPaths {
+		if p, err := exec.LookPath(path); err == nil {
+			return p, true
+		}
+	}
+	return "", false
+}
+
+// useVirtiofs checks if virtiofs should be used based on environment variable.
+// Returns (true, nil) if virtiofs is enabled and available.
+// Returns (false, nil) if virtiofs is not requested.
+// Returns (false, error) if virtiofs is explicitly requested but not available.
+func useVirtiofs() (bool, error) {
+	if envVal, ok := os.LookupEnv("QEMU_USE_VIRTIOFS"); ok {
+		if val, err := strconv.ParseBool(envVal); err == nil && val {
+			if _, available := isVirtiofsdAvailable(); available {
+				return true, nil
+			}
+			return false, fmt.Errorf("QEMU_USE_VIRTIOFS=1 but virtiofsd not found (checked %v, or set QEMU_VIRTIOFS_PATH)", virtiofsdSearchPaths)
+		}
+	}
+	return false, nil
+}
+
+// startVirtiofsd starts the virtiofsd daemon and returns its process.
+func startVirtiofsd(ctx context.Context, cfg *Config) (*exec.Cmd, error) {
+	log := clog.FromContext(ctx)
+
+	virtiofsdPath, ok := isVirtiofsdAvailable()
+	if !ok {
+		return nil, fmt.Errorf("virtiofsd not found")
+	}
+
+	// Remove stale socket if exists
+	os.Remove(cfg.VirtiofsdSocketPath)
+
+	args := []string{
+		"--socket-path=" + cfg.VirtiofsdSocketPath,
+		fmt.Sprintf("--thread-pool-size=%d", runtime.NumCPU()*2), // Parallel I/O
+		"-o", "source=" + cfg.CacheDir,
+		"-o", "cache=always", // Balance coherency and performance
+		"-o", "sandbox=namespace", // Use namespace sandbox (works without root)
+		"-o", "xattr", // Enable xattr support
+		"-o", "writeback", // Enable writeback caching for better write performance
+		"-o", "no_posix_lock",
+	}
+
+	log.Debugf("starting virtiofsd: %s %v", virtiofsdPath, args)
+
+	// #nosec G204 - virtiofsdPath is from known paths checked by isVirtiofsdAvailable
+	cmd := exec.CommandContext(ctx, virtiofsdPath, args...)
+
+	// Capture stderr for debugging
+	errPipe, err := cmd.StderrPipe()
+	if err != nil {
+		return nil, fmt.Errorf("failed to create stderr pipe: %w", err)
+	}
+
+	if err := cmd.Start(); err != nil {
+		return nil, fmt.Errorf("failed to start virtiofsd: %w", err)
+	}
+
+	// Log virtiofsd output in background
+	go func() {
+		scanner := bufio.NewScanner(errPipe)
+		for scanner.Scan() {
+			log.Debugf("virtiofsd: %s", scanner.Text())
+		}
+	}()
+
+	// Wait for socket to be created (max 5 seconds)
+	for range 50 {
+		if _, err := os.Stat(cfg.VirtiofsdSocketPath); err == nil {
+			log.Infof("virtiofsd started successfully, socket: %s", cfg.VirtiofsdSocketPath)
+			return cmd, nil
+		}
+		time.Sleep(100 * time.Millisecond)
+	}
+
+	// Socket not created, kill process and return error
+	_ = cmd.Process.Kill()
+	return nil, fmt.Errorf("virtiofsd socket not created within timeout")
+}
+
+// stopVirtiofsd gracefully stops the virtiofsd daemon.
+func stopVirtiofsd(ctx context.Context, cfg *Config) {
+	log := clog.FromContext(ctx)
+
+	if cfg.VirtiofsdPID > 0 {
+		log.Debugf("stopping virtiofsd (PID %d)", cfg.VirtiofsdPID)
+
+		// Send SIGTERM for graceful shutdown
+		if err := syscall.Kill(cfg.VirtiofsdPID, syscall.SIGTERM); err != nil {
+			log.Warnf("failed to send SIGTERM to virtiofsd: %v", err)
+			// Force kill if SIGTERM fails
+			_ = syscall.Kill(cfg.VirtiofsdPID, syscall.SIGKILL)
+		}
+
+		cfg.VirtiofsdPID = 0
+	}
+
+	// Clean up socket file
+	if cfg.VirtiofsdSocketPath != "" {
+		if err := os.Remove(cfg.VirtiofsdSocketPath); err != nil && !os.IsNotExist(err) {
+			log.Warnf("failed to remove virtiofsd socket: %v", err)
+		}
+		cfg.VirtiofsdSocketPath = ""
+	}
+}
+
 // zeroSensitiveFields zeros out sensitive cryptographic material from memory after it's no longer needed.
 func zeroSensitiveFields(ctx context.Context, cfg *Config) {
 	clog.FromContext(ctx).Debug("qemu: zeroing sensitive cryptographic material from memory")
@@ -1756,16 +1947,20 @@ func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 
 	cacheDir = filepath.Join(cacheDir, "melange-cpio")
 
+	// Include additional packages in cache filename to invalidate cache when they change
+	additionalPkgs := getAdditionalPackages(ctx)
+	cacheSuffix := getPackageCacheSuffix(additionalPkgs)
+
 	baseInitramfs := filepath.Join(
 		cacheDir,
-		"melange-guest.initramfs.cpio")
+		fmt.Sprintf("melange-guest%s.initramfs.cpio", cacheSuffix))
 	initramfsInfo, err := os.Stat(baseInitramfs)
 
 	// Check if we can use the cached base initramfs (less than 24h old)
 	useCache := err == nil && time.Since(initramfsInfo.ModTime()) < 24*time.Hour
 
 	if !useCache {
-		err = generateBaseInitramfs(ctx, cfg, baseInitramfs, cacheDir)
+		err = generateBaseInitramfs(ctx, cfg, baseInitramfs, cacheDir, additionalPkgs)
 		if err != nil {
 			return "", err
 		}
@@ -1776,7 +1971,7 @@ func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 
 // generateBaseInitramfs creates the base initramfs with microvm-init.
 // This is cached for performance as it doesn't change often.
-func generateBaseInitramfs(ctx context.Context, cfg *Config, initramfsPath, cacheDir string) error {
+func generateBaseInitramfs(ctx context.Context, cfg *Config, initramfsPath, cacheDir string, additionalPkgs []string) error {
 	clog.FromContext(ctx).Info("qemu: generating base initramfs")
 
 	err := os.MkdirAll(cacheDir, 0o755)
@@ -1784,14 +1979,17 @@ func generateBaseInitramfs(ctx context.Context, cfg *Config, initramfsPath, cach
 		return fmt.Errorf("unable to create dest directory: %w", err)
 	}
 
+	// Start with base packages and add any additional packages from environment
+	packages := make([]string, 0, len(additionalPkgs)+1)
+	packages = append(packages, "microvm-init")
+	packages = append(packages, additionalPkgs...)
+
 	spec := apko_types.ImageConfiguration{
 		Contents: apko_types.ImageContents{
 			BuildRepositories: []string{
 				"https://apk.cgr.dev/chainguard",
 			},
-			Packages: []string{
-				"microvm-init",
-			},
+			Packages: packages,
 		},
 	}
 	opts := []apko_build.Option{
@@ -1851,6 +2049,53 @@ func generateBaseInitramfs(ctx context.Context, cfg *Config, initramfsPath, cach
 	return nil
 }
 
+// getAdditionalPackages parses and validates the QEMU_ADDITIONAL_PACKAGES environment variable.
+// Returns a list of package names to add to the initramfs, or empty slice if none/invalid.
+func getAdditionalPackages(ctx context.Context) []string {
+	pkgEnv := os.Getenv("QEMU_ADDITIONAL_PACKAGES")
+	if pkgEnv == "" {
+		return nil
+	}
+
+	// Basic validation: check for suspicious characters that could cause injection
+	// Allow: alphanumeric, hyphens, underscores, commas, dots
+	if matched, _ := regexp.MatchString(`^[a-zA-Z0-9_,.-]+$`, pkgEnv); !matched {
+		clog.FromContext(ctx).Warnf("qemu: QEMU_ADDITIONAL_PACKAGES contains invalid characters, ignoring: %s", pkgEnv)
+		return nil
+	}
+
+	clog.FromContext(ctx).Infof("qemu: QEMU_ADDITIONAL_PACKAGES env set to %s, adding to initramfs", pkgEnv)
+
+	// Split comma-separated list and filter empty entries
+	var packages []string
+	for pkg := range strings.SplitSeq(pkgEnv, ",") {
+		pkg = strings.TrimSpace(pkg)
+		if pkg != "" {
+			packages = append(packages, pkg)
+		}
+	}
+
+	return packages
+}
+
+// getPackageCacheSuffix generates a deterministic cache suffix based on the package list.
+// Uses SHA256 hash (first 12 chars) to avoid collisions and keep filenames reasonable.
+// Returns empty string if packages list is empty.
+func getPackageCacheSuffix(packages []string) string {
+	if len(packages) == 0 {
+		return ""
+	}
+
+	// Join packages with commas to create a stable input string
+	pkgString := strings.Join(packages, ",")
+
+	// Generate SHA256 hash
+	hash := sha256.Sum256([]byte(pkgString))
+
+	// Use first 12 characters of hex encoding (like git short SHA)
+	return fmt.Sprintf("-%x", hash[:6])
+}
+
 // injectHostKeyIntoCpio creates a new initramfs by concatenating the base
 // initramfs with a secondary cpio containing the VM's SSH host key.
 // This allows us to use cached base initramfs while injecting fresh keys each time.
